#!/usr/bin/perl -w

#
# TODO: fill in missing module description
#
## @file
# Implementation of POS::LdapWorkstation

## @class
# Module responsible for workstation registration and update, writes data to LDAP.
# One instance for all workstation (only general data are stored in module, registration data
# are temporarily passed
package POS::LdapWorkstation;
use strict;


# version  of the package

our $VERSION = 0.1;

use Exporter;
use POS::Config;
use POS::Syslog;
use POS::LDAP;

our @ISA = qw( Exporter );

our @EXPORT = qw(init_branches );

## @EXPORT_OK;


#Programmcode: constructor

sub new
{
    my $package = shift;    ## packagename


    #my $config = POS::Config->new(); not needed?

    # Definition of object data, should be only general data common to all workstation instances
    my $data = {
        'STATIC' => undef,    ## hash of scLocation relevant data
        'CACHE'  => undef,    ## reference on cached data

        #        'WORKSTATION'     => undef,    ## name of the Workstation
        #        'MACADDRESS'      => undef,    ## MAC-Address of the NIC of the WS
        #        'IPADDR'       => undef,    ## IP-ADDRESS (dynamic or static)
        ## of a WORKSTATION
        #        'NETMASK'         => undef,    ## the Netmask, not really used in
        #        'WORKSTATIONLIST' => undef,    ## List of all Workstations in a
        ## Branch
        #        'WORKSTATIONHASH' => undef,    ## A Workstation-Hash
        #        'CLIENT'          => undef,    ## Object is used as Client,     - WTF?
        #        'LDAP'            => undef,    ## Ldap Connection
        #        'BRANCHBASE'        => undef,    ## LDAP-Base - dn of used scLocation
        #        'LDAPBASE'        => undef,    ## LDAP-Base - dn of used scLocation
        'ERROR' =>
          undef,    ## holds error msg and number if error occured and needs to be stored (e.g. to display on terminal)
                    #        'CACHE'            => undef,    ## Cached LDAP and other data
         #        'DEFAULTIP'      => '0.0.0.77'      # ip temporarily used if registering with EXT_DHCP and not having any
    };
    ## Here you can add other Variables


    set_var($data, @_);


    #    %$data=(%$data,@_);
    my $err = _init_data($data);
    if ($err) {
        log_alert_print("LdapWorkstation module init error, cannot continue");
        return undef;
    }
    # return the object-hash
    return bless($data);
} ## end sub new


#adds/(re)sets a variable of this object
sub set_var
{
    (my $self, my %data) = @_;
    %$self = (%$self, %data);
} ## end sub set_var


# reads and resets error of this object
# errdata hash naming must be compatible
# with names used in posLeaseMonitor.pm
# (this object is common for all workstations,
#  specific error data should be removed before
#  another workstation is checked/registered)
# TODO: some general error parsing module
sub get_error
{
    my ($self, $keep) = @_;
    my %errdata;
    $errdata{'CODE'}    = 0;
    $errdata{'MESSAGE'} = '';
    if (defined $self->{'ERROR'}) {
        $self->{'ERROR'} =~ /^(-?\d+):\s*(.*$)/;
        $errdata{'CODE'}    = $1;
        $errdata{'CODE'}    = -255 if !defined $errdata{'CODE'};
        $errdata{'MESSAGE'} = $2;
    }
    $self->{'ERROR'} = undef unless $keep;
    return \%errdata;
} ## end sub get_error


# now done in CACHE module
#initializes workstation object data relevant for workstation registration
# (ip ranges, network card...)
# not to be called from outside of this module
# TODO: cache here some more data from CACHE to simplify access and make resistant to CACHE change
sub _init_data
{
    (my $self) = @_;
    my $ntw_hr = $self->{'CACHE'}{'NETWORKCARDS'};
    # find proper NIC device
    if (keys %{$ntw_hr} == 0) {
        log_warn_printf("No scNetworkcard defined under \'%s\' this may be a problem..",
            $self->{'CACHE'}{'BRANCH'}{'SERVERCONTAINER'});
    }
    else {
        my $entry;
        if (keys %{$ntw_hr} > 1) {
            log_debug_printf('more than one NIC defined under "%s"', $self->{'CACHE'}{'BRANCH'}{'BRANCHSERVER'});
         # multiple NIC defined, find one with same IP as dns bind ip, this idea taken from original nlpos nic selecting
         # get dns ip
            my $dnsIp;
            $dnsIp = $self->{'CACHE'}{'SERVICES'}{'dns'}{'ipHostNumber'};
            $dnsIp =
              $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($self->{'STATIC'}{'BRANCHBASE'})->get_value('scDefaultGw')
              unless $dnsIp;
            # match with defined entries
            foreach (values %{$ntw_hr}) {
                if ($_->get_value('ipHostNumber') eq $dnsIp) {
                    $entry = $_;
                    last;
                }
            }
            if (not $entry) {
           # as a desparate measure use the first one, at least we will not fail but registration may not work correctly
           # OTOH we may just die to force user to fix their LDAP
                log_warn_print(
                    'Unable to find any network card for CR network! Taking first available. Validate LDAP!');
            }
        }
        $entry = (values %{$ntw_hr})[0] unless $entry;
        $self->{'ETH'} = $entry->get_value('scDevice');
        # check if ETH is virtual
        if (not grep(/$self->{'ETH'}/, `ls -1 /sys/class/net`)) {
            log_debug_printf('NIC %s is virtual, looking for bonded physical device', $self->{'ETH'});
            # it is, find bonded phy device to use
            my ($tmp) = `ip -o addr show` =~ qr([0-9]+: ([^ ]+).*$self->{'ETH'});
            if (not $tmp) {
                log_warn_printf(
'%s detected to be virtual device but no bonded physical device found. Registration may not work properly',
                    $self->{'ETH'}
                );
            }
            else {
                $self->{'ETH'} = $tmp;
            }
        }
        log_debug_printf('Using NIC %s (%s)', $self->{'ETH'}, $entry->get_value('ipHostNumber'));
    }

    my ($dyn_from, $dyn_to) = split(/,/, $self->{'CACHE'}{'BRANCH'}{'scLocation'}{'scDhcpRange'});
    $dyn_from =~ s/ //g;
    $dyn_to   =~ s/ //g;
    $self->{'DYN_FROM'} = $dyn_from;
    $self->{'DYN_TO'}   = $dyn_to;

    my ($fixed_from, $fixed_to) = split(/,/, $self->{'CACHE'}{'BRANCH'}{'scLocation'}{'scDhcpFixedRange'});
    $fixed_from =~ s/ //g;
    $fixed_to   =~ s/ //g;
    $self->{'FIXED_FROM'} = $fixed_from;
    $self->{'FIXED_TO'}   = $fixed_to;

    return 0;

} ## end sub _init_data


## testing the IP address
##  and creating IP associated data
##  errors are also generetaed inside
##  assuming having IP address and
##  all initialization data ready
## returns get_error()
## TODO: put this into posDnsDhcp.pm or LeaseMonitor.pm module
##       or some module containing other
##       checks used in 'gather_registration_data'
##      and assume ip is ok in this module
sub process_ip
{
    my ($self, $ip) = @_;
    ## testing the ip address sanity
    ##  and compatibility if no EXT_DHCP
    if (!$ip) {
        $self->{'ERROR'} = "-11: IP address is not defined";
        return $self->get_error();
    }
    ## Computing the Networknumber from the IP address
    ## Modul NetAddr::IP is required
    my $netmask;
    if ($self->{'STATIC'}{'EXTDHCP'}) {
        $netmask = '255.255.255.255';    #assume, just for ip sanity check
    }
    else {
        if ($self->{'CACHE'}{'BRANCH'}{'NETMASK'}) {
            $netmask = $self->{'CACHE'}{'BRANCH'}{'NETMASK'};
        }
        else {
            $self->{'ERROR'} = "-12: NETMASK is not defined";
            return $self->get_error();
        }
    }
    require NetAddr::IP;
    my $ip_test_obj = NetAddr::IP::Lite->new($ip, $netmask);

    if (!$ip_test_obj) {
        $self->{'ERROR'} = sprintf("-13: Can't create NetAddr::IP object, ip=%s is invalid?", $ip);
        return $self->get_error();
    }
    my $network_number = $ip_test_obj->network->addr();
    if (!$network_number) {
        $self->{'ERROR'} = sprintf("-14: Can't compute network-number for %s/%s", $ip, $netmask);
        return $self->get_error();
    }
    if (!$self->{'STATIC'}{'EXTDHCP'} && ($network_number ne $self->{'CACHE'}{'BRANCH'}{'NETNUM'})) {
        $self->{'ERROR'} = sprintf(
            "-15: Error, LDAP branch network-number (%s) differs from network-number derived from IP %s/%s",
            $self->{'CACHE'}{'BRANCH'}{'NETNUM'},
            $ip, $netmask
        );
        return $self->get_error();
    }
    #check $network_number to netnum?

    if (!$self->{'STATIC'}{'EXTDHCP'} && !$self->{'CACHE'}{'BRANCH'}{'IPMAP'}) {
    # check if IP address in dynamic OR static range, do not do this for ip mapping (responsibility of mapping function)
        my $isdyn = $self->_ip_in_dyn($ip);
        return $self->get_error() if !defined $isdyn;
        my $istat = $self->_ip_in_stat($ip);
        return $self->get_error() if !defined $istat;
        if (!$istat && !$isdyn) {
            $self->{'ERROR'} = sprintf("-18: IP address %s is neither in dynamic, nor in static pool, error.", $ip);
            return $self->get_error();
        }
    } ## end EXTDHCP=FALSE segment
    return $self->get_error();
} ## end sub process_ip


# helper procedure of process_ip
# basic data must be initialized
sub _ip_in_dyn
{
    my ($self, $ip) = @_;
    return $self->_ip_in_range($ip, $self->{'DYN_FROM'}, $self->{'DYN_TO'});
} ## end sub _ip_in_dyn

sub _ip_in_stat
{
    my ($self, $ip) = @_;
    return $self->_ip_in_range($ip, $self->{'FIXED_FROM'}, $self->{'FIXED_TO'});
} ## end sub _ip_in_stat

sub _ip_in_range
{
    (my $self, my $ip, my $r1, my $r2) = @_;
    require POS::IP;
    my $ip_addr = POS::IP->new($ip);
    if (!$ip_addr) {
        $self->{'ERROR'} = "-20: cannot create POS::IP object from $ip";
        return undef;
    }
    if ($ip_addr->is_in_range($r1, $r2)) {
        return 1;
    }
    else {
        return 0;
    }
} ## end sub _ip_in_range


#checks if ip is already used,
# by testing a response from it
sub _check_ip_used
{
    my ($self, $ip) = @_;

} ## end sub _check_ip_used


###############################################
###############################################
# add_workstation functions
sub _map_ws_name_default
{
    my $self               = shift;
    my $workstation_number = shift;

    my $ws_base_name = $self->{'CACHE'}{'BRANCH'}{'scLocation'}{'scWorkstationBaseName'};
    my $ws_num       = $self->{'CACHE'}{'BRANCH'}{'scLocation'}{'scEnumerationMask'};
    my $ws_num_len   = length($ws_num);

    if ($workstation_number =~ /^[0-9]+$/) {    #number
        return $ws_base_name . sprintf("%0*d", $ws_num_len, $workstation_number);
    }
    elsif ($workstation_number =~ /^[0-9.]+$/) {    # ip address
        $workstation_number =~ s/\./-/g;
        return $ws_base_name . '-' . $workstation_number;
    }
    return undef;
} ## end sub _map_ws_name_default

sub _map_ws_name_direct
{
    my $self = shift;
    my $id   = shift;
    my $name = $id;

    $name =~ s/\./-/g;
    $name =~ s/[^-0-9A-Za-z]//g;
    $name =~ s/^-*//;
    $name =~ s/-*$//;
    log_info_print("Using dns-compatible name '$name' for ID '$id'") if ($name ne $id);
    return $name;
} ## end sub _map_ws_name_direct

sub _map_ws_name_transform
{
    my $self   = shift;
    my $id     = shift;
    my $regexp = shift;

    if ($id =~ /$regexp/) {
        my $name = $1;
        my $orig = $1;

        $name =~ s/\./-/g;
        $name =~ s/[^-0-9A-Za-z]//g;
        $name =~ s/^-*//;
        $name =~ s/-*$//;
        log_info_print("Using dns-compatible name '$name' for '$orig'") if ($name ne $orig);
        return $name;
    }
    log_warn_printf("unable to get ws name from '$id' with regexp '$regexp'");
    return undef;
} ## end sub _map_ws_name_transform

sub _map_ws_ip_default
{
    my $self               = shift;
    my $workstation_number = shift;

    my $new_ip =
      _ws_number_to_ip($self, $workstation_number, $self->{'FIXED_FROM'}, $self->{'FIXED_TO'},
        $self->{'CACHE'}{'BRANCH'}{'NETNUM'});
    if (!$new_ip) {
        $self->{'ERROR'} = "-815: DHCP address outside of valid fixed range.";
        log_error_printf($self->{'ERROR'});
        return undef;
    }
    return $new_ip;
} ## end sub _map_ws_ip_default

sub _map_ws_ip_direct
{
    my $self = shift;
    my $id   = shift;
    if ($id =~ /^[0-9.]*[0-9]$/) {
        my $val = 0;
        for my $part (split(/\./, $id)) {
            $val = $val * 256 + $part;
        }
        if ($val == 0) {
            log_warn_printf("'$id' must be non-zero for use as an IP");
            return undef;
        }
        my $net_ip = new POS::IP($self->{'CACHE'}{'BRANCH'}{'NETNUM'});
        my $to_add = ($net_ip->intip() | $val) - $net_ip->intip();
        my $ip =
          NetAddr::IP::Lite->new($self->{'CACHE'}{'BRANCH'}{'NETNUM'}, $self->{'CACHE'}{'BRANCH'}{'NETMASK'}) + $to_add;
        return $ip->addr();
    }
    log_warn_printf("'$id' can't be converted to IP");
    return undef;
} ## end sub _map_ws_ip_direct

sub _map_ws_ip_transform
{
    my $self   = shift;
    my $id     = shift;
    my $regexp = shift;

    if ($id =~ /$regexp/) {
        return _map_ws_ip_direct($self, $1);
    }
    log_warn_printf("unable to get IP/ws_number from '$id' with regexp '$regexp'");
    return undef;
} ## end sub _map_ws_ip_transform

sub _get_free_ws_number_from_name
{
    my $self = shift;

    #try to keep the number first
    my $entry = $self->{'REGDAT'}{'scWorkstation'};
    if ($entry) {
        my $name         = $entry->get_value('cn');
        my $ws_base_name = $self->{'CACHE'}{'BRANCH'}{'scLocation'}{'scWorkstationBaseName'};
        my $ws_num       = $self->{'CACHE'}{'BRANCH'}{'scLocation'}{'scEnumerationMask'};
        my $ws_num_len   = length($ws_num);
        if ($name =~ /^$ws_base_name([0-9]+)$/) {
            my $workstation_number = $1;
            if (length($workstation_number) == $ws_num_len) {
                log_info_print "keeping original workstation number $workstation_number for MAC " .
                  $self->{'REGDAT'}{'MAC'};
                return $workstation_number;
            }
        }
    }

    #get first unused
    my $workstation_number = 1;
    while (1) {
        my $name = _map_ws_name_default($self, $workstation_number);
        return $workstation_number if (!defined $self->{'CACHE'}{'WS'}{'ALL'}{$name});
        $workstation_number++;
    }
} ## end sub _get_free_ws_number_from_name

sub _get_free_ws_number_from_ip_range
{
    my $self = shift;

    #try to keep the number first
    my $entry = $self->{'REGDAT'}{'scWorkstation'};
    if ($entry) {
        my $ip = $entry->get_value('ipHostNumber');
        my $workstation_number =
          _ip_to_ws_number($self, $ip, $self->{'FIXED_FROM'}, $self->{'FIXED_TO'},
            $self->{'CACHE'}{'BRANCH'}{'NETNUM'});
        if (defined $workstation_number) {
            log_info_print "keeping original workstation number $workstation_number for MAC " .
              $self->{'REGDAT'}{'MAC'};
            return $workstation_number;
        }

    }

    #get first unused
    my $workstation_number = 1;
    while (1) {
        my $ip =
          _ws_number_to_ip($self, $workstation_number, $self->{'FIXED_FROM'}, $self->{'FIXED_TO'},
            $self->{'CACHE'}{'BRANCH'}{'NETNUM'});
        return if (!defined $ip);                                                            # end of range
        return $workstation_number if (!defined $self->{'CACHE'}{'WS'}{'BY-IP-ALL'}{$ip});
        $workstation_number++;
    }
} ## end sub _get_free_ws_number_from_ip_range

sub _get_ws_name_and_ip
{
    my $self   = shift;
    my $id     = shift;
    my $use_ip = shift;

    $id = '' unless defined $id;

    my $scDnsMapFunc = $self->{'CACHE'}{'BRANCH'}{'scLocation'}{'scDnsMapFunc'};
    my $scIpMapFunc  = $self->{'CACHE'}{'BRANCH'}{'scLocation'}{'scIpMapFunc'};

    my $name;
    my $ip;

    if (!$scDnsMapFunc || $scDnsMapFunc =~ /^none$/i) {
        # do nothing, use fallback
    }
    elsif ($scDnsMapFunc =~ /^direct$/i) {
        $name = _map_ws_name_direct($self, $id);
    }
    elsif ($scDnsMapFunc =~ /^transform:(.*)$/i) {
        $name = _map_ws_name_transform($self, $id, $1);
    }
    else {
        log_error_print("invalid scDnsMapFunc value: $scDnsMapFunc");
    }

    if ($use_ip) {
        if (!$scIpMapFunc || $scIpMapFunc =~ /^none$/i) {
            # do nothing, use fallback
        }
        elsif ($scIpMapFunc =~ /^direct$/i) {
            $ip = _map_ws_ip_direct($self, $id);
        }
        elsif ($scIpMapFunc =~ /^transform:(.*)$/i) {
            $ip = _map_ws_ip_transform($self, $id, $1);
        }
        else {
            log_error_print("invalid scIpMapFunc value: $scIpMapFunc");
        }
    }

    # fallback : get an unused ws number and generate the name
    my $ws_num;
    if ($use_ip && !$ip) {
        $ws_num = _get_free_ws_number_from_ip_range($self);
        if (!$ws_num) {
            log_error_print("Can't get an unused IP - Fixed range is full");
            return;
        }
    }
    elsif (!$name) {
        # if the ip address was specified by id, use it for name too
        # otherwise fallback to the old numbering scheme
        $ws_num = ($ip) ? $ip : _get_free_ws_number_from_name($self);
        if (!$ws_num) {
            log_error_print("Can't get an unused workstation number");
            return;
        }
    }

    if (!$name) {
        $name = _map_ws_name_default($self, $ws_num);
    }

    if ($use_ip && !$ip) {
        $ip = _map_ws_ip_default($self, $ws_num);
        log_error_print("DHCP address outside of valid fixed range.") if !$ip;
    }
    return ($name, $ip);
} ## end sub _get_ws_name_and_ip


# assuming register data are correct (should have been tested before)
# all relevant data shoul be either in module's hash variables, or in
# $regdat_ref (expendable, no need to update, is only used here)
# (former workstation_update)
sub process_workstation
{
    #register new workstation into LDAP or update existing one
    # all ldap writes of posLeaseMonitor.pm should only happen here
    my ($self, $regdat_ref) = @_;
    my $ws_dn;
    $self->{'REGDAT'} = $regdat_ref;

    if (!$regdat_ref->{'MAC'}) {    #should not really happen under normal circumstances
        $self->{'ERROR'} = "-10: MAC address is not defined";
        log_alert_printf($self->{'ERROR'});
        return $self->get_error();
    }
    #if dump(all), delete update flag only
    if ($regdat_ref->{'FLAGS'}{'LDAPREG'} && !$regdat_ref->{'FLAGS'}{'WSUPDATE'}) {
        $ws_dn = $regdat_ref->{'scWorkstation'}->dn();
        $regdat_ref->{'WS_DN'} = $ws_dn;
        my $ret = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, replace => ['scConfigUpdate' => []]);
        if ($ret->code) {
            $self->{'ERROR'} = sprintf('LDAP-Error: Can\'t remove scConfigUpdate: Error=%s ', $ret->error);
            log_alert_printf($self->{'ERROR'});
        }
        return $self->get_error();
    }    #continue with update even for DUMP..?


    $ws_dn = $self->_update_workstation();
    return $self->get_error() if !defined $ws_dn;

    $regdat_ref->{'WS_DN'} = $ws_dn;

    $self->{'REGDAT'} = undef;    #remove temp data

    ##############
    #?
    ##    $regdat_ref->{'WSLDAPIP'} = $self->{'LDAPSERVER'}{'NEWIPADDRESS'};
    #########################################

    #continue with registration, code common with the update of an existing registration
    # TODO: see that --dumpall really works
    #$regdat_ref->{'scWorkstation'} exists in both cases

    # TODO write here using reliable ldap access functions (new in LDAP.pm module)

    #write scPosRegisterType
    my $hwtype = $regdat_ref->{'POS_HWTYPE'};
    if (defined $hwtype) {
        my $ret = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, replace => ['scPosRegisterType' => $hwtype]);
        if ($ret->code) {
            $self->{'ERROR'} = sprintf('LDAP-Error: Can\'t modify scPosRegisterType of \'%s\' to \'%s\': Error=%s ',
                $ws_dn, $hwtype, $ret->error);
            log_warn_printf($self->{'ERROR'});
            return $self->get_error();
        }
    }
    #write scPosRegisterBiosVersion
    my $biosinfo = $regdat_ref->{'HWTYPESEL'}{'DATA'}{'HWBIOS'};
    #if ($biosinfo = $self->{'HWINFO'}{$mac}{'HWBIOS'})) {
    if (defined $biosinfo) {
        my $ret = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, replace => ['scPosRegisterBiosVersion' => $biosinfo]);
        if ($ret->code) {
            $self->{'ERROR'} =
              sprintf('LDAP-Error: Can\'t modify scPosRegisterBiosVersion of \'%s\' to \'%s\': Error=%s ',
                $ws_dn, $biosinfo, $ret->error);
            log_warn_printf($self->{'ERROR'});
            return $self->get_error();
        }
    }

    #write scRefPcDn  (cashregister)
    my $crdn = $regdat_ref->{'scCashRegister'}->dn();
    my $ret = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, replace => ['scRefPcDn' => $crdn]);
    if ($ret->code) {
        $self->{'ERROR'} =
          sprintf('LDAP-Error: Can\'t modify scRefPcDn of \'%s\' to \'%s\': Error=%s ', $ws_dn, $crdn, $ret->error);
        log_warn_printf($self->{'ERROR'});
        return $self->get_error();
    }

    #write if rolebased
    my $rolebased = $regdat_ref->{'FLAGS'}{'ROLEBASED'};
    if (defined $rolebased) {
        my $rolebasedstr = 'TRUE';
        $rolebasedstr = 'FALSE' if ($rolebased == 0);
        my $ret = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, replace => ['scRoleBased' => $rolebasedstr]);
        if ($ret->code) {
            $self->{'ERROR'} = sprintf('LDAP-Error: Can\'t modify scRoleBased of \'%s\' to \'%s\': Error=%s ',
                $ws_dn, $rolebasedstr, $ret->error);
            log_warn_printf($self->{'ERROR'});
            return $self->get_error();
        }
    }

    #write scRoleDn  (role used)
    my $roledn = $regdat_ref->{'POS_ROLE'};
    if (defined $roledn) {
        my $ret = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, replace => ['scRoleDn' => $roledn]);
        if ($ret->code) {
            $self->{'ERROR'} =
              sprintf('LDAP-Error: Can\'t modify scRoleDn of \'%s\' to \'%s\': Error=%s ', $ws_dn, $roledn,
                $ret->error);
            log_warn_printf($self->{'ERROR'});
            return $self->get_error();
        }
    }

    #write scId
    my $id = $regdat_ref->{'POS_ID'};
    if (defined $id) {
        if ($id ne '') {
            my $ret = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, replace => ['scId' => $id]);
        }
        else
        { #empty id means delede attribute (and not write '' as the attribute value, which is possible, but not correct here)
            my $ret = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, replace => ['scId' => []]);
        }
        if ($ret->code) {
            $self->{'ERROR'} =
              sprintf('LDAP-Error: Can\'t write (scId: %s) to %s: Error=%s ', $id, $ws_dn, $ret->error);
            log_warn_printf($self->{'ERROR'});
            return $self->get_error();
        }
    }
    #already read from LDAP if needed
    #else {    # if POS_ID not defined, return current state (=> retrun current state always, unless error above)
    #    $regdat_ref->{'POS_ID'} = $regdat_ref->{'scWorkstation'}->get_value('scId');
    #}


    #update cache
    my $ws_obj  = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($ws_dn);
    my $wscn    = $ws_obj->get_value('cn');
    my $wsip    = $ws_obj->get_value('ipHostNumber');
    my @ws_macs = $ws_obj->get_value('macAddress');
    $self->{'CACHE'}{'WS'}{'ACT'}{$wscn} = $ws_obj;
    delete $self->{'CACHE'}{'WS'}{'DEACT'}{$wscn};
    $self->{'CACHE'}{'WS'}{'BY-IP-ALL'}{$wsip} = $ws_obj;
    foreach my $ws_mac (@ws_macs) {
        $self->{'CACHE'}{'WS'}{'BY-MAC'}{$ws_mac} = $ws_obj;
    }
    #TODO: update ID list


    ##unless dumpall,  explicit image refernce will be removed, so the machine registers as expected
    ##(if dumpall, we probably want custom config or at least can get one, if this would apply always, there woui
    ##do this if rolebased only???
    ##if (!$self->{'DUMPALL'}) {
    #$self->{'LDAP'}->modify($ws_dn, replace => { 'scPosImageDn' => '' });
    ##}

    return $self->get_error();

} ## end sub process_workstation


#
#!!!!!!!!!!!make this cmpatible with new data structures


# use all available infomation to update or register the machine.
# not all inforamtion may be provided, error only if really needed
# IPADDR may not be defined, use DEFAULTIP if needed
sub _update_workstation
{
    my ($self)  = @_;
    my $regdatr = $self->{'REGDAT'};
    my $ip      = $regdatr->{'IPADDR'};
    $ip = $regdatr->{'WSLDAPIP'} if !defined $ip;
    $ip = '' if !defined $ip;

    my $id_s = '';
    $id_s = ' ID:' . $regdatr->{'POS_ID'} if defined $regdatr->{'POS_ID'};
    log_debug_printf(
        "Registering MAC:%s IP:%s$id_s EXTDHCP:%s",
        $self->{'REGDAT'}{'MAC'},
        $ip, $self->{'STATIC'}{'EXTDHCP'}
    );


    ## ip should have already been tested to be safe
    ## testing the parameter IPADDRESS
    ## and creating IP associated data
    ## errors are also generated inside
    #$regdatr->process_ip($self->{'IPADDR'});
    #my $new_reg = $self->{'CACHE'}{'EXTDHCP'} ?
    #              $self->process_ip_extDHCP($self->{'IPADDR'}) :
    #              $self->process_ip($self->{'IPADDR'});

    #TODO: simplify whole process
    # note on workstations deleted from ldap but having still entry in dhcp.conf. May indicate other problem too..
    if ($self->_ip_in_stat($ip) && !$self->{'STATIC'}{'EXTDHCP'} && $regdatr->{'FLAGS'}{'NEWREG'}) {
        log_warn_printf(
"Note, %s is in the static range, workstation seems to already have an entry in dhcp.conf, has it been registered before and then manually deleted from LDAP? Registration can continue.",
            $ip
        );
    }

    ##check
    ## the IPADDRESS is in the dynamic or static range (reused)
    my ($workstation_name, $new_ip) = _get_ws_name_and_ip($self, $regdatr->{'POS_ID'}, !$self->{'STATIC'}{'EXTDHCP'});
    $new_ip = $ip if ($self->{'STATIC'}{'EXTDHCP'});    #ignore if ip external

    return if (!$workstation_name || !$new_ip);

    if ($ip ne $new_ip &&
        system('arping', '-q', '-f', '-w1', '-I', $self->{'ETH'}, $new_ip) == 0) {
        $self->{'ERROR'} = "-31: IP $new_ip is used by another machine";
        log_error_print($self->{'ERROR'});
        return undef;
    }

    log_info_printf("Try to add %s as $workstation_name with IP: $new_ip", $self->{'REGDAT'}{'MAC'});

    invalidate_workstation($self, $self->{'REGDAT'}{'POS_MAC'}, $workstation_name);
    ## connecting to ldap

    ## configuring the data to send
    my $dn = "cn=" . $workstation_name . ",";
    $dn .= $self->{'STATIC'}{'BRANCHBASE'};
    log_info_print "adding: " . $dn;

    ## maybe it is better to use some CONFIG_DATA
    my @myobjectclass = ();
    push(@myobjectclass, "scWorkstation");
    push(@myobjectclass, "top");

    my $result = $self->{'CACHE'}{'LDAPCACHE'}->add(
        $dn,
        attr => [
            cn          => $workstation_name,
            objectClass => [@myobjectclass],
            #         macAddress   => uc($self->{'REGDAT'}{'MAC'}),
            macAddress   => $self->{'REGDAT'}{'POS_MAC'},
            ipHostNumber => $new_ip,
        ]
    );

    if ($result->code) {
        #re-use existing scWorkstation object
        $result =
          $self->{'CACHE'}{'LDAPCACHE'}
          ->modify($dn, replace => [macAddress => $self->{'REGDAT'}{'POS_MAC'}, ipHostNumber => $new_ip]);
    }
    elsif ($result->code) {
        $self->{'ERROR'} = "-17: Error adding Workstation: $workstation_name " . $result->error;
        #				$ldap->unbind();
        return undef;
    }
    $self->{'REGDAT'}{'WSLDAPIP'} = $new_ip;
    # all ok
    return $dn;

    #    }    #fi of '$self->process($self->{'IPADDRESS'}' condition
    #   return undef;

} ## end sub _update_workstation

#sub add_workstation
#{
#    my $self = shift;
#    ## undefining MACADDRESS and IPADDRESS
#    $self->{'MACADDRESS'} = undef;
#    $self->{'IPADDRESS'}  = undef;
#    $self->{'POS_ID'}  = undef;
#    $self->{'CACHE'}{'EXTDHCP'}  = undef;
#    while (@_) {
#        my $field = shift;
#        if (@_) {
#            if ($field eq 'MACADDRESS' ||
#                $field eq 'IPADDRESS' ||
#                $field eq 'POS_ID' ||
#                $field eq 'EXTDHCP') {
#                $self->{$field} = shift;
#                if (!$self->{$field}) {
#                    $self->{$field} = "";
#                }
#            }
#        }
#    }
#    log_debug_printf("Registering MAC:%s IP:%s ID:%s EXTDHCP:%s", $self->{'MACADDRESS'}, $self->{'IPADDRESS'}, $self->{'POS_ID'}, $self->{'CACHE'}{'EXTDHCP'});
#
#    ## testing the parameter MACADDRESS
#    if (!$self->{'MACADDRESS'}) {
#        $self->{'ERROR'}{'CODE'} = -10;
#        $self->{'ERROR'}{'MESSAGE'}  = "MACADDRESS is not defined";
#        return undef;
#    }
#
#    ## testing the parameter IPADDRESS
#    ## and creating IP associated data
#    ## errors are also generetaed inside
#    my $new_reg = $self->{'CACHE'}{'EXTDHCP'} ?
#                  $self->process_ip_extDHCP($self->{'IPADDRESS'}) :
#                  $self->process_ip($self->{'IPADDRESS'});
#    if (1) {
#
#        ## the IPADDRESS is in the dynamic range
#        my ($workstation_name, $new_ip) = _get_ws_name_and_ip($self, $self->{'POS_ID'}, !$self->{'CACHE'}{'EXTDHCP'});
#        $new_ip = $self->{'IPADDRESS'} if ($self->{'CACHE'}{'EXTDHCP'});
#
#        return if (!$workstation_name || !$new_ip);
#
#        if ($self->{'IPADDRESS'} ne $new_ip &&
#            system('arping', '-q', '-f', '-w1',
#                             '-I', $self->{'BRANCH'}{'scNetworkcard'}{'scDevice'},
#                             $new_ip) == 0) {
#            $self->{'ERROR'}{'CODE'} = -17;
#            $self->{'ERROR'}{'MESSAGE'}  = "IP $new_ip is used by another machine";
#            log_error_print($self->{'ERROR'}{'MESSAGE'});
#            return undef;
#        }
#
#        log_info_printf("Try to add %s as $workstation_name with IP: $new_ip", $self->{'MACADDRESS'});
#        $self->{'NEWIPADDRESS'} = $new_ip;
#
#        invalidate_workstation($self, $self->{'MACADDRESS'}, $workstation_name);
#        ## connecting to ldap
#
#        ## configuring the data to send
#        my $dn = "cn=" . $workstation_name . ",";
#        $dn .= $self->{'BRANCH'}{'scLocation'}{'dn'};
#        log_info_print "adding: " . $dn;
#
#        ## maybe it is better to use some CONFIG_DATA
#        my @myobjectclass = ();
#        push(@myobjectclass, "scWorkstation");
#        push(@myobjectclass, "top");
#
#        my $result = $self->{'LDAP'}->add($dn,
#            attr => [
#                cn           => $workstation_name,
#                objectClass  => [@myobjectclass],
#                macAddress   => uc($self->{'MACADDRESS'}),
#                ipHostNumber => $new_ip,
#            ]);
#
#        if ($result->code) {
#            #re-use existing scWorkstation object
#            $result = $self->{'LDAP'}->modify($dn,
#                replace => [
#                    macAddress   => uc($self->{'MACADDRESS'}),
#                    ipHostNumber => $new_ip
#                ]);
#        }
#        elsif ($result->code) {
#            $self->{'ERROR'}{'CODE'} = -17;
#            $self->{'ERROR'}{'MESSAGE'}  = "Error adding Workstation: $workstation_name " . $result->error;
#            #				$ldap->unbind();
#            return undef;
#        }
#
#        # all ok
#        $self->{'ERROR'}{'CODE'} = 0;
#        $self->{'ERROR'}{'MESSAGE'}  = "";
#        return $dn;
#    }    #fi of '$self->process($self->{'IPADDRESS'}' condition
#    return undef;
#} ## end sub add_workstation


# invalidate all old occurences of registration MAC(s) if new name
sub invalidate_workstation
{
    my $self     = shift;
    my $mac_ar   = shift;
    my $new_name = shift;

    foreach my $mac (@{$mac_ar}) {
# TODO:  also solve invalid case when already more than one WS has same MAC
        my $entry = $self->{'CACHE'}{'WS'}{'BY-MAC'}{$mac} if exists $self->{'CACHE'}{'WS'}{'BY-MAC'}{$mac};
#      $ldap->search(base => $self->{'STATIC'}{'BRANCHBASE'}, filter => "macAddress=$mac_address", attrs => ['cn', 'macAddress']);

        if (!$entry) {
            log_warn_print("No active workstation found with MAC=$mac");
            next;
        }

        my $dn = $entry->dn;
        my $cn = $entry->get_value('cn');

        #update cache by ip (even for $cn eq $new_name)
        my $ldapip = $entry->get_value('ipHostNumber');
        delete $self->{'CACHE'}{'WS'}{'BY-IP-ALL'}{$ldapip} if defined $ldapip;

        if ($new_name ne $cn) {
            log_info_print("invalidating MAC $mac in Workstation $cn");
            my $result2 = $self->{'CACHE'}{'LDAPCACHE'}->modify($dn, delete => { 'macAddress' => $mac });
            if ($result2->code) {
                log_error_print("Error invalidating MAC $mac in Workstation $cn : " . $result2->error);
            }
            #update cache
            if (!defined $entry->get_value('macAddress')) {    #no mac left, update as deactivated
                $self->{'CACHE'}{'WS'}{'DEACT'}{$cn} = $entry;
                delete $self->{'CACHE'}{'WS'}{'ACT'}{$cn};
            }
            delete $self->{'CACHE'}{'WS'}{'BY-MAC'}{$mac};
        }
    }

} ## end sub invalidate_workstation

sub _ip_to_ws_number
{
    my ($self, $ip_addr, $fixed_from, $fixed_to, $network_number) = @_;
    if (!$ip_addr || !$fixed_from) {
        $self->{'ERROR'} = "-41: ip-address or network number is not defined";
        return undef;
    }

    my $start_ip = new POS::IP($fixed_from);
    if (!$start_ip) {
        $self->{'ERROR'} = "-42: fixed_from ($fixed_from): invalid ip address.";
        return undef;
    }
    my $end_ip = new POS::IP($fixed_to);
    if (!$end_ip) {
        $self->{'ERROR'} = "-43: fixed_to ($fixed_to): invalid ip address.";
        return undef;
    }
    my $ip = new POS::IP($ip_addr);
    if (!$ip) {
        $self->{'ERROR'} = "-44: ip number ($ip_addr): invalid ip address.";
        return undef;
    }
    my $num  = sprintf("%d", $ip->intip() - $start_ip->intip()) + 1;
    my $last = sprintf("%d", $end_ip->intip() - $start_ip->intip()) + 1;
    $num = undef if $num <= 0 || $num > $last;
    return $num;
} ## end sub _ip_to_ws_number

sub _ws_number_to_ip
{
    my ($self, $workstation_number, $fixed_from, $fixed_to, $network_number) = @_;

    # split the fixed dhcp from boundary into its pieces
    my @ffrom = split(/\./, $fixed_from);
    for (my $i = 0; $i < 4; $i++) {
        $ffrom[$i] = $ffrom[$i] + 0;
    }
    my @fto = split(/\./, $fixed_to);
    for (my $i = 0; $i < 4; $i++) {
        $fto[$i] = $fto[$i] + 0;
    }

    # don't want floating point in the / and %
    use integer;

    # need to decrement workstation number because we want to use
    # the low end network number
    my $offset = ($workstation_number - 1);
    # Loop through calculating ip from the workstation number
    # NOTE: This will now vend 255 and 0 if valid.
    for (my $i = 3; $i >= 0 and ($offset + 0) > 0; $i--) {
        my $mod = $ffrom[$i] + $offset;
        $offset = $mod / 256;

        $ffrom[$i] = $mod % 256;
    }

    # Now compare the address we have with the upper end of
    # the fixed range. This should ultimately issue a Syslog
    # to indicate that they have walked out of their ip space.
    if ($ffrom[0] < $fto[0] ||
        (   $ffrom[0] == $fto[0] && $ffrom[1] < $fto[1] ||
            ($ffrom[1] == $fto[1] && $ffrom[2] < $fto[2] ||
                ($ffrom[2] == $fto[2] && $ffrom[3] <= $fto[3]))
        )
      ) {
        # This one is ok
        return join(".", @ffrom);
    }

    return undef;
} ## end sub _ws_number_to_ip


1;

